function [I,sr,dout,dqout,yout,aiout,biout,tdout,trout,alout] = ...
		 adpcm(sl, N, format);
%ADPCM	Adaptive differential pulse code modulation.
%	[I, SR] = ADPCM( SL, N) encodes the PCM coded input signal SL
%	with N kbit/s adaptive differential pulse code modulation (ADPCM) 
%	where N is selected from 40, 32, 24, or 16 kbit/s. 
%	If N is not specified, the bit rate is set to N = 32 kbit/s.
%
%	By default, the input signal SL is assumed to be represented with
%	14-bit signed-magnitude uniform PCM, i.e., its amplitude range is
%	-8191 <= sl <= 8191. This input format is also selected by
%	ADPCM(SL, N, 'lin'). If the input signal is 8-bit logarithmic PCM
%	(mu-law in accordance with LIN2MU), ADPCM( SL, N, 'mu') has to
%	be used.
%
% 	The first output is I, the encoder output signal which represents 
%	the bit stream over the channel. The samples of I are from the 
%	discrete alphabet -2^(N/8-1), ..., +2^(N/8-1). The second (optional) 
% 	output is the reconstructed speech signal SR (scaled approximately 
%	between -8191 and +8191) which would be the output of the corres-
%	ponding ADPCM decoder in the case of error-free transmission of I.
%	
%	If more than two outputs are specified, a whole list of
%	diagnostic outputs (internal variables of the algorithm) becomes 
%	available:
%
%		[I,SR,D,DQ,Y,AI,BI,TD,TR,AL] = ADPCM( SL, N);
%
%	with the following interpretation:
%
%		D	prediction difference signal
%		DQ	quatized prediction difference signal
%		Y	scale factor for adaptive quantizer 
%		AI	recursive predictor coefficients (2-column matrix)
%		BI	non-recursive predictor coefficients (6-column matrix)
%		TD	tone detector signal
%		TR	transient detector signal
%		AL	speed control for scale factor adaptation
%
%	Note that this latter option is VERY memory intensive as all
%	variables are vectors/matrices of the size of the input signal.
%
%	This function uses also APUPDATE, AIUPDATE, BIUPDATE.
% 	The decoder is provided by I_ADPCM.
%
%  	References:	
%	The algorithm follows closely the ITU-T (former CCITT)
%	recommendation G.726 "40, 32, 24, 16 kbit/s Adaptive Differential
% 	Pulse Code Modulation (ADPCM)". The recommendation specifies
%	various fixed-point formats for the internal variables which are
%	not simulated in this MATLAB implementation. However, most of the 
%	variable names etc. are consistent with the recommendation.
 
% 1995-05-16, Martin Kummernecker
% 1995-08-30, Gernot Kubin (g.kubin@ieee.org)



% input argument checks, default bit-rate and format setting
if (nargin < 1) | (nargin > 3)
	disp('ADPCM: 1 to 3 input variables required!')
	help adpcm, return
end
if nargin == 1
	N = 32;
    disp('***** N = 32 kbit/s *****');
end
a = N==[40 32 24 16];
if (sum(a) ~= 1)
   N = 32;
   disp('***** N = 32 kbit/s *****');
end;
if nargin <= 2
	format = 'lin';
end;
if isstr(format) ~= 1
	error('ADPCM: Input format must be string variable!')
elseif strcmp(format,'mu') 
	sl = 8191*mu2lin(sl);
elseif ~strcmp(format, 'lin')
	error('ADPCM: wrong input format')
end
if min(size(sl)) ~= 1
   error('ADPCM: Input signal must be a row or column vector only');
end;
sl = sl(:);		        % force input signal to be a column vector
if max(abs(sl)) > 8191
   disp('ADPCM warning: there can be clipping');
   disp(['You should use the  amplitude range of ',...
	     '-8191 <= SL <= 8191 for the input signal SL!']);
end;
if max(sl) <= 1
   disp('ADPCM warning: there can be underflow');
   disp(['You should use the amplitude range of ',...
	     '-8191 <= SL <= 8191 for the input signal SL!']);
end;

	
I  = zeros(size(sl));	% allocate output memory (channel symbols)
sr = zeros(size(sl));	% allocate output memory (reconstructed speech)
if (nargout > 2)		% allocate memory for diagnostic outputs
	dout  = zeros(size(sl));
	dqout = zeros(size(sl));
	yout  = zeros(size(sl));
	aiout = zeros(length(sl),2);
	biout = zeros(length(sl),6);
	tdout = zeros(size(sl));
	trout = zeros(size(sl));
	alout = zeros(size(sl));
end

% define quantizer tables and scale-factor adaptation tables
if N == 40
   h_fi   = fliplr([6 6 5 4 3 2 1 1 1 1 1 0 0 0 0 0]);
   h_wi   = fliplr([43.50 33.06 27.50 22.38 17.50 13.69, ...
		    11.19 8.81 6.25 3.63 2.56 2.50 2.44 1.50 0.88 0.88]);
   h_iadq = fliplr([4.42 4.21 4.02 3.81 3.59 3.35 3.09 2.80 2.48, ...
			2.14 1.75 1.32 0.81 0.22 -0.52 -inf]);
   h_qan  = fliplr([4.31 4.12 3.91 3.70 3.47 3.22 2.95 2.64 2.32, ...
		    1.95 1.54 1.08 0.52 -0.13 -0.96 -inf]);
end;
if N == 32
   h_fi   = fliplr([7 3 1 1 1 0 0 0]);
   h_wi   = fliplr([70.13 22.19 12.38 7.00 4.00 2.56 1.13 -0.75]);
   h_iadq = fliplr([3.32 2.91 2.52 2.13 1.66 1.05 0.031 -inf]);
   h_qan  = fliplr([3.12 2.72 2.34 1.91 1.38 0.62 -0.98 -inf]);
end;
if N == 24
   h_fi   = fliplr([7 2 1 0]);
   h_wi   = fliplr([36.38 8.56 1.88 -0.25]);
   h_iadq = fliplr([2.91 2.13 1.05 -inf]);
   h_qan  = fliplr([2.58 1.70 0.06 -inf]);
end;
if N == 16
   h_fi   = fliplr([7 0]);
   h_wi   = fliplr([27.44 -1.38]);
   h_iadq = fliplr([2.85 0.91]);
   h_qan  = fliplr([2.04 -inf]);
end;

% numerical constants (leakage, step sizes etc.)
c1  = 0.125;		% 2^-3
c2  = 0.9375;		% 1-2^-4
c3  = 0.03125;		% 2^-5
c4  = 0.96875;		% 1-2^-5
c5  = 0.015625;	% 2^-6
c6  = 0.984375;	% 1-2^-6
c7  = 0.0078125;	% 2^-7
c8  = 0.9921875;	% 1-2^-7
c9  = 0.01171875;	% 3*2^-8
c10 = 0.99609375;	% 1-2^-8
c11 = 0.998046875;	% 1-2^-9

% initialize internal variables
yu=1.06; yl=0; dms=0; dml=0; ap=0;
dq = zeros(1,7);	% difference signal vector [dq(k),dq(k-1),...,dq(k-6)]
ai = zeros(1,2);	% recursive predictor coefficients [a1(k),a2(k)]
bi = zeros(1,6);	% non-recursive predictor coefficients 
					   % [b1(k),b2(k),...,b6(k)]
p  = zeros(1,3);	% surrogate vector for recursive predictor
					   % adaptation [p(k),p(k-1),p(k-2)]
srv = zeros(1,2);	% reconstructed-speech vector [sr(k),sr(k-1)]

% loop through all signal samples
for k = 1:length(sl)

	% difference signal computation and quantization
   sez  = bi * dq(1:6)';	   % signal prediction, non-recursive part
   se   = ai * srv' + sez;	% signal prediction, recursive part
   d    = sl(k) - se;		   % difference signal computation
   al   = min(ap, 1);			% hard-limit speed control parameter
   y    = al*yu + (1-al)*yl;	% update quantizer scale factor
	x    = log2(abs(d)) - y;	% scale difference signal in log domain
	I(k) = sign(d) * (sum(h_qan <= x)-1); 	% flash quantizer

	% inverse quantization and signal reconstruction
   dqh  = sign(I(k)) * 2^(h_iadq(abs(I(k))+1)+y);	% inverse quant./log
   dq   = [dqh dq(1:6)];		% shift difference signal vector
   p    = [dq(1)+sez, p(1:2)]; % shift surrogate vector
  	sr(k)= se + dq(1);			% reconstruct speech sample
   srv  = [sr(k), srv(1)]; 	% shift reconstructed-speech vector 

	% adaptation of predictor coefficients
   ai   = aiupdate(ai,p,c9,c10,c7,c8,c2);
   bi   = biupdate(bi,dq,N,c7,c10,c11);	

	% quantizer scale factor adaptation
	WI   = h_wi(abs(I(k))+1);
	yu   = c4*y + c3*WI;		% fast (unlocked) scale factor
	yu   = max(min(yu,10.00),1.06);
   yl   = c6*yl + c5*yu;		% slow (locked) scale factor

	% tone and transition detection
   td   = ai(2) < -0.71875;		% partial band signal detection
   tr   = td & (abs(dq(1)) > 24*2^yl);
   if tr==1
      ai = zeros(1,2);
      bi = zeros(1,6);
   end;

	% quantizer adaptation speed control
	FI   = h_fi(abs(I(k))+1);
   dml  = c8*dml + c7*FI;		% long term average of F[I(k)]
   dms  = c4*dms + c3*FI;		% short term average of F[I(k)]
   ap   = apupdate(ap,dms,dml,y,td,tr,c1,c2);	% speed control parameter


	% write diagnostic outputs
	if (nargout > 2)
		dout(k)    = d;
		dqout(k)   = dqh;
		yout(k)    = y;
		aiout(k,:) = ai;
		biout(k,:) = bi;
		tdout(k)   = td;
		trout(k)   = tr;
		alout(k)   = al;
	end
end;
